from math import log
from termcolor import colored
from re import search, IGNORECASE

from Nemesis.lib.Globals import Color
from Nemesis.lib.Globals import base64char, hexchar, dom_sources_regex, dom_sinks_regex
from Nemesis.lib.Globals import url_regex, subdomain_regex, path_regex, single_path_regex
from Nemesis.lib.Globals import web_services_regex, custom_regex_sensitive, custom_regex_insensitive

def banner():
    b = '\x1b[1m\x1b[31m    _   __                         _     \n   / | / /__  ____ ___  ___  _____(_)____\n  /  |/ / _ \\/ __ `__ \\/ _ \\/ ___/ / ___/\n / /|  /  __/ / / / / /  __(__  ) (__  ) \n/_/ |_/\\___/_/ /_/ /_/\\___/____/_/____/  \n                                         \n\x1b[0m'
    print(b)

def starter(argv):
    from sys import stdin
    if argv.banner:
        banner()
        exit()
    if not argv.wordlist:
        if not argv.url:
            if not argv.stdin:
                print(f"{Color.bad} Use --help")
                exit()
            else:
                return (line.rstrip('\n').strip(' ') for line in stdin.read().split('\n') if line)
        else:
            return [argv.url.strip(' ')]
    else:
        return (line.rstrip('\n').strip(' ') for line in open(argv.wordlist) if line)

def shannon_entropy(data, iterator):
    if not data:
        return 0
    entropy = 0
    for val in iterator:
        p_x = float(data.count(val))/len(data)
        if p_x > 0:
            entropy += - p_x * log(p_x, 2)
    return float(entropy)

def reduce_string(word: str, args = []) -> str:
    if not word:
        return ""
    word = word.rstrip('//').rstrip(';')
    for arg in args:
        word = word.strip(arg)
    return word

def dom_source_extract(line: str) -> tuple:
    output_list = ()
    for dom_source in dom_sources_regex:
        if search(dom_source, line, IGNORECASE):
            mline = search(dom_source, line, IGNORECASE).group()
            output_list = (mline, 'dom_source_match')
    return output_list

def dom_sink_extract(line: str) -> tuple:
    output_list = ()
    for dom_sink in dom_sinks_regex:
        if search(dom_sink, line, IGNORECASE):
            mline = search(dom_sink, line, IGNORECASE).group()
            output_list = (mline, 'dom_sink_match')
    return output_list

def subdomain_extract(line: str, domain: str) -> tuple:
    output_list = ()
    subdomain = subdomain_regex(domain)
    if search(subdomain, line, IGNORECASE):
        mline = search(subdomain, line, IGNORECASE).group()
        if not mline == domain:
            output_list = (mline, 'subdomain_match')
    return output_list

def url_extract(line: str) -> tuple:
    output_list = ()
    for web_service in web_services_regex:
        if search(web_service, line):
            mline = search(url_regex, line).group()
            output_list = (mline, 'web_services_match')
    if search(url_regex, line):
        mline = search(url_regex, line).group()
        output_list = (mline, 'url_match')
        #if unender(unurler(line), '/') == self.argv.domain:
        #    return ()
    return output_list

def path_extract(line: str) -> tuple:
    output_list = ()
    if search(path_regex, line):
        mline = search(path_regex, line).group()
        output_list = (mline, 'path_match')
    elif search(single_path_regex, line):
        mline = reduce_string(search(single_path_regex, line).group(), args=['"', "'"])
        output_list = (mline, 'single_path_match')
    return output_list

def shannon_extract(line: str) -> tuple:
    output_list = ()
    entropy = float(4)
    for word in line.split(' '):
        if len(word) > 5:
            if shannon_entropy(word, base64char) > entropy or shannon_entropy(word, hexchar) > entropy:
                word = reduce_string(word, args=['"', "'"])
                output_list = (word, 'shannon_entropy_match')
    return output_list

def custom_extract(line: str) -> tuple:
    output_list = ()
    for custom in custom_regex_sensitive:
        if search(custom, line):
            mline = search(custom, line).group()
            output_list = (mline, 'custom_regex')
    for custom in custom_regex_insensitive:
        if search(custom, line, IGNORECASE):
            mline = search(custom, line, IGNORECASE).group()
            output_list = (mline, 'custom_regex')
    return output_list

def link_extract(line: str, domain = "") -> tuple:
    # rename output_list to output_tuple
    output_list = ()
    if not line or line.startswith('#'):
        return output_list
    for web_service in web_services_regex:
        if search(web_service, line):
            mline = search(web_service, line).group()
            output_list = (mline, 'web_services_match')
            return output_list
    # url_regex over subdomain_regex, might supress
    if search(url_regex, line):
        mline = search(url_regex, line).group()
        output_list = (mline, 'url_match')
    elif search(path_regex, line):
        mline = search(path_regex, line).group()
        output_list = (mline, 'path_match')
    elif search(single_path_regex, line):
        mline = search(single_path_regex, line).group()
        output_list = (mline, 'single_path_match')
    #elif search(experimental_path_regex, line):
    #    mline = search(experimental_path_regex, 'experimental_path_match')
    #    output_list = (mline, 'experimental_path_match')
    elif domain and search(subdomain_regex(domain), line):
        mline = search(subdomain_regex(domain).group(), line)
        if not mline == domain:
            output_list = (mline, 'subdomain_match')
    return output_list

def pretty_print(line: str, match_type_tuple: tuple):
    if not match_type_tuple: return
    match, match_type = match_type_tuple
    match_type = match_type.split('_match')[0].replace('_', ' ')
    split_beginning, split_end = line.split(match)[0], line.split(match)[-1]
    colored_line = split_beginning + colored(match, color='red', attrs=['bold']) + split_end
    print(f"{Color.good} Found: {colored_line} (Match: {match_type})")



from requests import get
from bs4 import Comment
from jsbeautifier import beautify
from urllib.parse import unquote

class Engine:
    def __init__(self):
        pass

    def find_href(self, s):
        l =  [unquote(t['href']) for t in s.find_all(href = True) if t]
        return l

    def find_img_src(self, s):
        l = [unquote(i['src']) for i in s.find_all('image', {'src': True})]
        return l

    def find_comment(self, s) -> set:
        c = set(s.find_all(string=lambda text: isinstance(text, Comment)))
        return c

    def find_hidden_input(self, s):
        p = []
        l = s.find_all('input', {'type': 'hidden', 'name': True})
        for i in l:
            if i.has_attr('value'):
                p.append(i['name'] + ':' + i['value'])
            else:
                p.append(i['name'] + ":")
        return p

    def find_script_src(self, s):
        l = [unquote(st['src']) for st in s.find_all('script', {'src': True })]
        return l

    def find_script_code(self, s):
        e = [beautify(st.string).split('\n') for st in s.find_all('script', {'src': False })]
        return e

    def html_source_return(self, u):
        data = []
        try:
            data = get(u).text
        except Exception as E:
            print(E, E.__class__)
        return data

    def js_source_return(self, u):
        data = []
        try:
            data = beautify(get(u).text).split('\n')
        except Exception as E:
            print(E, E.__class__)
        return data



from termcolor import colored
from traceback import print_exc
from re import search, IGNORECASE
from urllib.parse import urlparse
from bs4 import BeautifulSoup

from Nemesis.lib.Engine import Engine
from Nemesis.lib.Globals import Color
from Nemesis.lib.PathFunctions import urler
from Nemesis.lib.Functions import link_extract
from Nemesis.lib.Functions import dom_source_extract, dom_sink_extract, path_extract
from Nemesis.lib.Functions import subdomain_extract, custom_extract, shannon_extract
from Nemesis.lib.Functions import pretty_print

engine = Engine()
def extract_url(url):
    unparsed_url = urler(url)
    parsed_url = urlparse(unparsed_url)
    domain, path, query = parsed_url.netloc, parsed_url.path, parsed_url.query
    is_js_url = path.endswith('.js')
    output_list = []
    if is_js_url:
        js_code = engine.js_source_return(unparsed_url)
    else:
        s = BeautifulSoup(engine.html_source_return(unparsed_url), 'html.parser')
        js_scripts = engine.find_script_code(s)

    for js_code in js_scripts:
        for js_line in js_code:
            line = js_line.strip(' ').rstrip('{').rstrip(' ').lstrip('}').lstrip(' ')
            l = dom_source_extract(line)
            if l:
                output_list.append(l)
                pretty_print(line, l)
                continue
            l = dom_sink_extract(line)
            if l:
                output_list.append(l)
                pretty_print(line, l)
                continue
            l = link_extract(line)
            if l:
                output_list.append(l)
                pretty_print(line, l)
                continue
            #l = path_extract(line)
            #if l:
            #    output_list.append(l)
            #    continue
            #l = subdomain_extract(line)
            #if l:
            #    output_list.append(l)
            #    continue
            l = custom_extract(line)
            if l:
                output_list.append(l)
                pretty_print(line, l)
                continue
            l = shannon_extract(line)
            if l:
                output_list.append(l)
                pretty_print(line, l)
                continue
    if not path or not is_js_url:
        print(url)
        html_dict = {
            'url': unparsed_url,
            'links': engine.find_href(s),
            'img_links': engine.find_img_src(s),
            'script_links': engine.find_script_src(s),
            'comments': engine.find_comment(s),
            'hidden_parameters': engine.find_hidden_input(s),
            }

        if html_dict['links']:
            print(f"{Color.information} General links:")
            for link in html_dict['links']:
                l = link_extract(link)
                pretty_print(link, l)
        if html_dict['img_links']:
            print(f"{Color.information} Image links:")
            for link in html_dict['img_links']:
                l = link_extract(link)
                pretty_print(link, l)
        if html_dict['script_links']:
            print(f"{Color.information} Exline scripts sources:")
            for link in html_dict['script_links']:
                l = link_extract(link)
                pretty_print(link, l)
        if html_dict['comments']:
            print(f"{Color.information} Comments:")
            for comment in html_dict['comments']:
                print(f"{Color.good} Found: {comment}")
        if html_dict['hidden_parameters']:
            print(f"{Color.information} Hidden parameters:")
            h = "&".join([hp.replace(":", "=") for hp in html_dict['hidden_parameters']])
            #h = hidden_parameter.replace(":", "=")
            print(f"{Color.good} Found: {colored(h, color='red', attrs=['bold'])}")



from Nemesis.lib.Color import Color

# General
# Matches url but not direct path
url_regex = "((http|https)\:\/\/)+[a-zA-Z0-9\.\/\?\:@\-_=#]+\.([a-zA-Z]){2,6}([a-zA-Z0-9\.\&\/\?\:@\-_=#])*"
single_path_regex = """('|"|\(|\))(\/){1}[a-zA-Z0-9-_]+(\/)?('|"|\(|\))"""
path_regex = "([a-zA-Z0-9]+\.[a-zA-Z0-9]{3,6})?\/(([0-9a-zA-Z+.-]+)([\/&| ])){1,30}([a-zA-Z0-9]+(\.[a-zA-Z0-9]*)?)?(\?|;)?([a-zA-Z\[\]&=]*)?"
experimental_path_regex = "([a-zA-Z0-9]+\.[a-zA-Z0-9]{3,6})?(\/)?(([0-9a-zA-Z+.-]+)([\/&| ])){1,30}([a-zA-Z0-9]+(\.[a-zA-Z0-9]*)?)?(\?|;)?([a-zA-Z\[\]&=]*)?"
# path should match a/anything /anything/ or /anything? but not /anything
subdomain_regex = lambda subdomain: '(.*\.)?{}(\.)?'.format(subdomain)

dom_sources_regex = [
'document.url',
'document.documenturi',
'Document.URLUnencoded',
'Document.baseURI',
'Location.href',
'Location.search',
'Location.hash',
'Location.pathname',
'Document.cookie',
'Document.referrer',
'Window.name',
'History.pushState',
'History.replaceState',
'LocalStorage',
'SessionStorage',
'window.location',
'document.location'
]

dom_sinks_regex = [
'eval',
'setTimeout',
'setInterval',
'setImmediate',
'execScript',
'cyrpto.generateCRMFRequest',
'ScriptElement',
'ScriptElement.src',
'ScriptElement.text',
'ScriptElement.textContent',
'ScriptElement.innerText',
'document.write',
'document.writeln',
]
custom_regex_not = [
        ''
]

custom_regex_insensitive = [
'secret',
'admin',
]

custom_regex_sensitive = [
'sourceMappingURL',
]

web_services_regex = [
'([0-9a-zA-Z-.]*s3[a-zA-Z0-9-.]*\.?amazonaws\.com\/?[a-zA-Z-.]*)',
'([0-9a-zA-Z-.]*?storage\.googleapis\.com\/?[a-zA-Z-.]*)',
'([0-9a-zA-Z-.]*?digitaloceanspaces\.com\/?[a-zA-Z-.]*)',
'([0-9a-zA-Z-.]*?blob\.core\.windows\.net\/?[a-zA-Z-.]*)',
]

hexchar = "1234567890abcdefABCDEF"
base64char = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
Color = Color()
